// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// QR decomposition with Householder
//
// Calling Sequence :
// qr(A)
//
// Arguments :
// A : DD matrix
//
// Return :
// Q : (m,m) orthogonal matrix
// R : (m,n) upper triangular matrix
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 
// *****************************************************************
function [Q,R] = %dd_qr(A)
  [m,n]=size(A.hi);
  Q = tlist(['dd','hi','lo'],eye(m,m),zeros(m,m));
  R = A;
  for k=0:n-1
    v = tlist(['dd','hi','lo'],zeros(m-k,1),zeros(m-k,1));
    v(1:m-k,1) = R(k+1:m,k+1);
    
    a = (-1.0*sign(v.hi(1,1)))* norm(v,2);
    v(1,1) = v(1,1) - a;
    
    tmp = 2.0/ddip(v,v)
    R(k+1:m,k+1:n) = R(k+1:m,k+1:n) - tmp*(v*(v'*R(k+1:m,k+1:n)));
    Q(k+1:m,:) = Q(k+1:m,:) - tmp*(v*(v'*Q(k+1:m,:)));
    clear v,tmp;
  end
  Q = Q';
endfunction

