#include <math.h> 


//******************************************************************************************//
// four basic arithmetic of double-double 
//  - addition
//  - subtraction
//  - multiplication
//  - division
//
// last modified : 2012/02/03
//
//******************************************************************************************//


//--------------------------------------------------------------------------------------------
//
// double-double addition
//
// args
// ahi alo, bhi, blo : double numbers
// ahi + alo = dd nubmer
// *m : row
// *n : column
// return (chi,clo) for dd number
//
// routines
// dd_a_dd : dd - dd addition
// dd_a_s  : dd - double addition
// s_a_dd  : double - dd addition
//
//--------------------------------------------------------------------------------------------

//dd addition
//dd + dd
int dd_a_dd(double *chi, double *clo, double *ahi, double *alo, double *bhi, double *blo, int *m, int *n)
{
	double sh,sl,eh,el,v;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		sh = ahi[i] + bhi[i];
		v = sh - ahi[i];
		eh = (ahi[i] - (sh - v)) + (bhi[i] - v);

        sl = alo[i] + blo[i];
		v = sl - alo[i];
		el = (alo[i] - (sl - v)) + (blo[i] - v);
        
		eh = eh + sl;
		
        v = sh + eh;
        eh = eh + (v - sh);
        
        eh = eh + el;
        
        chi[i] = v + eh;
		clo[i] = eh - (chi[i] - v);
	}
	return(0);
}

//dd addition
//dd + d
int dd_a_s(double *chi, double *clo, double *ahi, double *alo, double *b, double *m, double *n)
{
	double sh,eh,v;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		sh = ahi[i] + b[i];
		v = sh - ahi[i];
		eh = (ahi[i] - (sh - v)) + (b[i] - v);
		eh = eh + alo[i];
		chi[i] = sh + eh;
		clo[i] = eh - (chi[i] - sh);
	}
	return(0);
}

//dd addition
//d + dd
int s_a_dd(double *chi, double *clo, double *a, double *bhi, double *blo, double *m, double *n)
{
	double sh,eh,v;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		sh = a[i] + bhi[i];
		v = sh - a[i];
		eh = (a[i] - (sh - v)) + (bhi[i] - v);
		eh = eh + blo[i];
		chi[i] = sh + eh;
		clo[i] = eh - (chi[i] - sh);
	}
	return(0);
}


//--------------------------------------------------------------------------------------------
//
// double-double subtraction
//
// args
// ahi alo, bhi, blo : double numbers
// ahi + alo = dd nubmer
// *m : row
// *n : column
// return (chi,clo) for dd number
//
// routines
// dd_s_dd : dd - dd subtraction
// dd_s_s  : dd - double subtraction
// s_s_dd  : double - dd subtraction
//
//--------------------------------------------------------------------------------------------

//dd subtraction
//dd - dd
int dd_s_dd(double *chi, double *clo, double *ahi, double *alo, double *bhi, double *blo, double *m, double *n)
{
	double sh,sl,eh,el,v;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		bhi[i] = -bhi[i];
		blo[i] = -blo[i];
		
		sh = ahi[i] + bhi[i];
		v = sh - ahi[i];
		eh = (ahi[i] - (sh - v)) + (bhi[i] - v);
        
        sl = alo[i] + blo[i];
		v = sl - alo[i];
		el = (alo[i] - (sl - v)) + (blo[i] - v);
        
		eh = eh + sl;
		
        v = sh + eh;
        eh = eh +(v - sh);
        
        eh = eh + el;
        
        chi[i] = v + eh;
		clo[i] = eh - (chi[i] - v);
	}
	return(0);
}

//dd subtraction
//dd - d
int dd_s_s(double *chi, double *clo, double *ahi, double *alo, double *b, double *m, double *n)
{
	double sh,eh,v;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		b[i] = -b[i];
		
		sh = ahi[i] + b[i];
		v = sh - ahi[i];
		eh = (ahi[i] - (sh - v)) + (b[i] - v);
		eh = eh + alo[i];
		chi[i] = sh + eh;
		clo[i] = eh - (chi[i] - sh);
	}
	return(0);
}

//dd subtraction
//d - dd
int s_s_dd(double *chi, double *clo, double *a, double *bhi, double *blo, double *m, double *n)
{
	double sh,eh,v;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		bhi[i] = -bhi[i];
		blo[i] = -blo[i];
		sh = a[i] + bhi[i];
		v = sh - a[i];
		eh = (a[i] - (sh - v)) + (bhi[i] - v);
		eh = eh + blo[i];
		chi[i] = sh + eh;
		clo[i] = eh - (chi[i] - sh);
	}
	return(0);
}


//--------------------------------------------------------------------------------------------
//
// double-double multiplication
//
// args
// ahi alo, bhi, blo : double numbers
// ahi + alo = dd nubmer
// *m : row
// *n : column
// return (chi,clo) for dd number
//
// routines
// dd_m_dd : dd - dd multiplication
// dd_m_s  : dd - double multiplication
// s_m_dd  : double - dd multiplication
// scalar_multi : dd - dd multiplication (scalar * vector or matrix)
// mix_scalar_multi_1 : dd * d (scalar * vector or matrix) or d * dd (vector or matrix * scalar)
// mix_scalar_multi_2 : d * dd (scalar * vector or matrix) or dd * d (vector or matrix * scalar)
//
//--------------------------------------------------------------------------------------------

//dd multiplication
//dd * dd
int dd_m_dd(double *chi, double *clo, double *ahi, double *alo, double *bhi, double *blo, int *m, int *n, int *l)
{
	int i,j,k;
	double t,ah,al,bh,bl,p1,p2,chi_n,clo_n;
	
	for (i = 0 ; i < *m ; i++) {
		for (j = 0 ; j < *n ; j++) {	
			chi[(*m)*j+i] = 0.0;
			clo[(*m)*j+i] = 0.0;
			for (k = 0 ; k < *l ; k++) {
				//multiplication
				t = 134217729 * ahi[(*m)*k+i];
				ah = t - (t - ahi[(*m)*k+i]);
				al = ahi[(*m)*k+i] - ah;
				
				t = 134217729 * bhi[(*l)*j+k];
				bh = t - (t - bhi[(*l)*j+k]);
				bl = bhi[(*l)*j+k] - bh;
				
				p1 = ahi[(*m)*k+i] * bhi[(*l)*j+k];
				p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
				p2 = p2 + ahi[(*m)*k+i] * blo[(*l)*j+k];
    			p2 = p2 + alo[(*m)*k+i] * bhi[(*l)*j+k];
				
				chi_n = p1 + p2;
				clo_n = p2 - (chi_n - p1);
				
				//addition
				ah = chi[(*m)*j+i] + chi_n;
				t = ah - chi[(*m)*j+i];
				bh = (chi[(*m)*j+i] - (ah - t)) + (chi_n - t);
				
				al = clo[(*m)*j+i] + clo_n;
				bh = bh + al;
				chi[(*m)*j+i] = ah + bh;
				clo[(*m)*j+i] = bh - (chi[(*m)*j+i] - ah);
			}
		}
	}
	return(0);
}

//dd multiplication
//dd * d
int dd_m_s(double *chi, double *clo, double *ahi, double *alo, double *b, int *m, int *n, int *l)
{
	int i,j,k;
	double t,ah,al,bh,bl,p1,p2,chi_n,clo_n;
	
	for (i = 0; i < (*m); i++) {
		for (j = 0; j < (*n); j++) {
			chi[(*m)*j+i] = 0.0;
			clo[(*m)*j+i] = 0.0;
			for (k = 0; k < (*l); k++) {
				//multiplication
				t = 134217729 * ahi[(*m)*k+i];
				ah = t - (t - ahi[(*m)*k+i]);
				al = ahi[(*m)*k+i] - ah;
				
				t = 134217729 * b[(*l)*j+k];
				bh = t - (t - b[(*l)*j+k]);
				bl = b[(*l)*j+k] - bh;
				
				p1 = ahi[(*m)*k+i] * b[(*l)*j+k];
				p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
    			p2 = p2 + alo[(*m)*k+i] * b[(*l)*j+k];
				
				chi_n = p1 + p2;
				clo_n = p2 - (chi_n - p1);
				
				//addition
				ah = chi[(*m)*j+i] + chi_n;
				t = ah - chi[(*m)*j+i];
				bh = (chi[(*m)*j+i] - (ah - t)) + (chi_n - t);
				
				al = clo[(*m)*j+i] + clo_n;
				bh = bh + al;
				chi[(*m)*j+i] = ah + bh;
				clo[(*m)*j+i] = bh - (chi[(*m)*j+i] - ah);
			}
		}
	}
	return(0);
}

//dd multiplication
//d * dd
int s_m_dd(double *chi, double *clo, double *a, double *bhi, double *blo, int *m, int *n, int *l)
{
	int i,j,k;
	double t,ah,al,bh,bl,p1,p2,chi_n,clo_n,sh,eh,sl,v;
	
	for (i = 0; i < (*m); i++) {
		for (j = 0; j < (*n); j++) {
			
			chi[(*m)*j+i] = 0.0;
			clo[(*m)*j+i] = 0.0;
			
			for (k = 0; k < (*l); k++) {
				//multiplication
				t = 134217729 * a[(*m)*k+i];
				ah = t - (t - a[(*m)*k+i]);
				al = a[(*m)*k+i] - ah;
				
				t = 134217729 * bhi[(*l)*j+k];
				bh = t - (t - bhi[(*l)*j+k]);
				bl = bhi[(*l)*j+k] - bh;
				
				p1 = a[(*m)*k+i] * bhi[(*l)*j+k];
				p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
				p2 = p2 + a[(*m)*k+i] * blo[(*l)*j+k];
				
				chi_n = p1 + p2;
				clo_n = p2 - (chi_n - p1);
				
				//addition
				sh = chi[(*m)*j+i] + chi_n;
				v = sh - chi[(*m)*j+i];
				eh = (chi[(*m)*j+i] - (sh - v)) + (chi_n - v);
				
				sl = clo[(*m)*j+i] + clo_n;
				eh = eh + sl;
				chi[(*m)*j+i] = sh + eh;
				clo[(*m)*j+i] = eh - (chi[(*m)*j+i] - sh);
			}
		}
	}
	return(0);
}

//dd multiplication
//dd * dd (scalar * vector or matrix)
int scalar_multi(double *chi, double *clo, double *ahi, double *alo, double *bhi, double *blo, int *m, int *n)
{
	int i;
	double t,ah,al,bh,bl,p1,p2;
	
	t = 134217729 * ahi[0];
	ah = t - (t - ahi[0]);
	al = ahi[0] - ah;
	
	for(i = 0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		
		//multiplication
		t = 134217729 * bhi[i];
		bh = t - (t - bhi[i]);
		bl = bhi[i] - bh;
		
		p1 = ahi[0] * bhi[i];
		p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
		p2 = p2 + ahi[0] * blo[i];
		p2 = p2 + alo[0] * bhi[i];
		
		chi[i] = p1 + p2;
		clo[i] = p2 - (chi[i] - p1);
	}
	return(0);
}

//dd multiplication
//dd * d (scalar * vector or matrix) or d * dd (vector or matrix * scalar)
int mix_scalar_multi_1(double *chi, double *clo, double *ahi, double *alo, double *b, int *m, int *n)
{
	int i;
	double t,ah,al,bh,bl,p1,p2;
	
	t = 134217729 * ahi[0];
	ah = t - (t - ahi[0]);
	al = ahi[0] - ah;
	
	for(i = 0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		
		//multiplication
		t = 134217729 * b[i];
		bh = t - (t - b[i]);
		bl = b[i] - bh;
		
		p1 = ahi[0] * b[i];
		p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
		p2 = p2 + alo[0] * b[i];
		
		chi[i] = p1 + p2;
		clo[i] = p2 - (chi[i] - p1);
	}
	return(0);
}

//dd multiplication
//d * dd (scalar * vector or matrix) or dd * d (vector or matrix * scalar)
int mix_scalar_multi_2(double *chi, double *clo, double *a, double *bhi, double *blo, int *m, int *n)
{
	int i;
	double t,ah,al,bh,bl,p1,p2;
	
	t = 134217729 * a[0];
	ah = t - (t - a[0]);
	al = a[0] - ah;
	
	for(i = 0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		
		//multiplication
		t = 134217729 * bhi[i];
		bh = t - (t - bhi[i]);
		bl = bhi[i] - bh;
		
		p1 = a[0] * bhi[i];
		p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
		p2 = p2 + a[0] * blo[i];
		
		chi[i] = p1 + p2;
		clo[i] = p2 - (chi[i] - p1);
	}
	return(0);
}


//--------------------------------------------------------------------------------------------
//
// double-double division
//
// args
// ahi alo, bhi, blo : double numbers
// ahi + alo = dd nubmer
// *m : row
// *n : column
// return (chi,clo) for dd number
//
// routines
// dd_r_dd : dd - dd division
// dd_r_s  : dd - double division
// s_r_dd  : double - dd division
//
//--------------------------------------------------------------------------------------------

//dd division
//dd / dd
int dd_r_dd(double *chi, double *clo, double *ahi, double *alo, double *bhi, double *blo, int *m, int *n)
{
	double c,cc,ah,al,bh,bl,p1,p2;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		
		c = ahi[i]/bhi[0];
		p1 = c*bhi[0];
		cc = 134217729 * c;
		ah = cc - (cc - c);
		al = c - ah;
		cc = 134217729 * bhi[0];
		bh = cc - (cc - bhi[0]);
		bl = bhi[0] - bh;
		p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
		
		cc = (ahi[i] - p1 - p2 + alo[i] - c*blo[0]) / bhi[0];
		
		chi[i] = c + cc;
		clo[i] = cc - (chi[i] - c);
	}
	return(0);
}

//dd division
//dd / d
int dd_r_s(double *chi, double *clo, double *ahi, double *alo, double *b, int *m, int *n)
{
	double c,cc,ah,al,bh,bl,p1,p2;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		
		c = ahi[i]/b[0];
		p1 = c*b[0];
		cc = 134217729 * c;
		ah = cc - (cc - c);
		al = c - ah;
		cc = 134217729 * b[0];
		bh = cc - (cc - b[0]);
		bl = b[0] - bh;
		p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
		
		cc = (ahi[i] - p1 - p2 + alo[i]) / b[0];
		
		chi[i] = c + cc;
		clo[i] = cc - (chi[i] - c);
	}
	return(0);
}

//dd division
//d / dd
int s_r_dd(double *chi, double *clo, double *a, double *bhi, double *blo, int *m, int *n)
{
	double c,cc,ah,al,bh,bl,p1,p2;
	int i;
	for(i=0; i<(*m)*(*n); i++) {
		chi[i] = 0.0;
		clo[i] = 0.0;
		
		c = a[i]/bhi[0];
		p1 = c*bhi[0];
		cc = 134217729 * c;
		ah = cc - (cc - c);
		al = c - ah;
		cc = 134217729 * bhi[0];
		bh = cc - (cc - bhi[0]);
		bl = bhi[0] - bh;
		p2 = ((ah * bh - p1) + ah * bl + al * bh) + al * bl;
		
		cc = (a[i] - p1 - p2 - c*blo[0]) / bhi[0];
		
		chi[i] = c + cc;
		clo[i] = cc - (chi[i] - c);
	}
	return(0);
}

