// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// QR decomposition with modified Gram-Schmidt orthonormalization
//
// Calling Sequence :
// ddgqr(A)
//
// Arguments :
// A : DD matrix
//
// Return : 
// Q : m,n orthogonal matrix
// R : n~n upper triangular matrix
//
// make dll file with Visual C++2010 (ddnorm_two,dd_m_dd)
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 
// *****************************************************************
function [Q,R] = ddgqr(A)
  [m,n]=size(A.hi);
  Q = A;
  R = tlist(['dd','hi','lo'],zeros(n,n),zeros(n,n));
  z = call('ddnorm_two',Q(:,1).hi,2,'d',Q(:,1).lo,3,'d',m,4,'i','out',[2,1],1,'d');
  R(1,1) = tlist(['dd','hi','lo'],z(1),z(2));
  Q(:,1) = Q(:,1) / R(1,1);
  for i=2:n
    q = Q(:,i);
    [c,cc] = call('dd_m_dd',(Q(:,1:i-1).hi)',3,'d',(Q(:,1:i-1).lo)',4,'d',q.hi,5,'d',q.lo,6,'d',i-1,7,'i',m,8,'i',m,9,'i','out',[i-1,1],1,'d',[i-1,1],2,'d');
    R(1:i-1,i) = tlist(['dd','hi','lo'],c,cc);
    for j=1:i-1
      q = q - R(j,i)*Q(:,j);
    end
    z = call('ddnorm_two',q.hi,2,'d',q.lo,3,'d',m,4,'i','out',[2,1],1,'d');
    R(i,i) = tlist(['dd','hi','lo'],z(1),z(2));
    Q(:,i) = q / R(i,i);
  end
endfunction
