// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// N-th power of the DD number
//
// Calling Sequence :
// ddpow(a,n)
//
// Arguments :
// a : DD number
// n : integer
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 2011/11/01 (bug:function name)
// *****************************************************************
function b = ddpow(a,n)
  //
  absn = abs(n);
  if(a.hi == 0) then
    if(n > 0) then
      b = tlist(['dd','hi','lo'],0,0);
    else
      error(27) //Divide by zero...
    end
  else
    if(absn == 0) then
      b = tlist(['dd','hi','lo'],1,0);
    elseif(absn == 1) then
      b = a;
    elseif(absn == 2) then
      b = a*a;
    else
      // Determine the least integer MN such that 2^MN > MN
      mn = int(log2(exp(1))*log(absn) + 1.0 + 1.0D-14);
      s0 = a;
      b = tlist(['dd','hi','lo'],1,0);
      kn = absn;
      // Compute B^N using binary rule for exponentiation
      for j=1:mn
        kk = int(kn / 2);
        if(kn ~= 2.0*kk) then
          s1 = b * s0;
          b = s1;
        end
        kn = kk;
        if(j < mn) then
          s1 = s0 * s0;
          s0 = s1;
        end
      end
    end   
    // Compute reciprocal if N is negative
    if(n < 0) then
      b = 1.0 / b;
    end
  end
endfunction

