// ====================================================================
// Copyright (C) 2011 - Tsubasa Saito
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// display DD variable
//
// Calling Sequence :
// ddprint(a)
//
// Arguments :
// a : dd number (applied only scalar) 
//
// This function is based on Bailey's DDFUN90
// http://crd.lbl.gov/~dhbailey/mpdist/
//
// ***TODO:apply for matrix
//
// Authors :
// Copyright (C) 2011 - Tsubasa Saito
// last updating : 
// *****************************************************************
function c = ddprint(a)
  ln = 32;
  ib = zeros(ln,1);
  digits = ['0';'1';'2';'3';'4';'5';'6';'7';'8';'9'];
  f = tlist(['dd','hi','lo'],10,0);
  s = tlist(['dd','hi','lo'],0,0);
  
  // Determine exact power of ten for exponent.
  if(a.hi ~= 0) then
    t1 = log10(abs(a.hi));
    if(t1 >= 0) then
      nx = int(t1);
    else
      nx = int(t1) - 1;
    end
    s = a /(ddpow(f,nx));
    
    if(s.hi < 0) then
      s.hi = -s.hi;
      s.lo = -s.lo;
    end
    
    // If we didn't quite get it exactly right,multiply or divide by 10 to fix.
    i = 0;
    
    if(1 <= s.hi & s.hi < 10) then
      bool = %F;
    else
      bool = %T;
    end
    
    while(s.hi < 1 | s.hi >= 10 | bool == %T)
      i = i + 1;
      if(s.hi < 1) then
        nx = nx - 1;
        s = s * 10;
        if(i > 3) then
          bool = %F;
        end
      elseif(s.hi >= 10) then
        nx = nx + 1;
        s = s / 10;
      end
    end
  else
    nx = 0;
  end
  
  // conpute digits
  
  for i=1:ln
    ib(i) = int(s.hi);
    s = (s - ib(i)) * 10;
  end
  
  // fix negative digits
  
  for i=ln:-1:2
    if(ib(i) < 0) then
      ib(i) = ib(i) + 10;
      ib(i-1) = ib(i-1) - 1;
    end
  end
  
  if(ib(1) < 0) then
    disp('ddprint : negative leading digit')
    c = 0;
    abort;
  end
  
  // round
  
//  if(ib(ln) >= 5) then
//    ib(ln-1) = ib(ln-1) + 1;
    
//    for i=ln-1:-1:2
//      if(ib(i) == 10) then
//        ib(i) = 0;
//        ib(i-1) = ib(i-1) + 1;
//      end
//    end
    
//    if(ib(1) == 10) then
//      ib(1) = 1;
//      nx = nx + 1;
//    end
//  end
  
  // insert digit characters in ib
  
  c = '  ';
  if(a.hi >= 0) then
    c = c + ' ';
  else
    c = c + '-';
  end
  
  c = c + digits(ib(1)+1);
  c = c + '.';
  
  for i=2:ln-1
    c = c + digits(ib(i)+1);
  end
  
  // insert exponent
  
  c = c + 'E';
  is = sign(nx);
  d1 = abs(nx);
  
  if(is < 0) then
    c = c + '-';
  end
  ca = '';
  for i=1:4
    d2 = int(d1/10);
    k = 1 + (d1 - 10*d2);
    d1 = d2;
    ca(i,1) = digits(k);
    if(d1 == 0) then
      break;
    end
  end
  for j=i:-1:1
    c = c + ca(j,1);
  end
endfunction

