#include <math.h> 

inline void fast_two_sum(double *s, double *e, double a, double b);
inline void two_sum(double *s, double *e, double a, double b);
inline void three_sum(double *d, double *e, double *f, double a, double b, double c);
inline void three_sum2(double *d, double *e, double a, double b, double c);
inline void two_prod(double *p, double *e, double a, double b);
inline void sqr(double *p, double *e, double a);
inline void renorm(double *b0, double *b1, double *b2, double *b3, double a0, double a1, double a2, double a3, double a4);
inline void qd_add_s(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b);
inline void qd_add_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3);
inline void qd_sub_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3);
inline void s_mul_qd(double *c0, double *c1, double *c2, double *c3, double b, double a0, double a1, double a2, double a3);
inline void qd_mul_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3);
inline void qd_div_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3);
inline void qd_sqr(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3);
inline void qdsqrt_scl(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3);

// including files
//   qdgauss    : gaussian elimination
//   qdlu       : LU decomposition
//   qdinv      : inverse matrix
//   qdip       : inner product
//   qdnorm_two : vector 2-norm
//   qdsum      : vector summation


// Gaussian elimination with pivoting
// For solving systems of linear equations with QD arithmetic
int qdgauss(double *x0, double *x1, double *x2, double *x3, double *a0, double *a1, double *a2, double *a3, double *b0, double *b1, double *b2, double *b3, int *n)
{
	int i,j,k,max_i;
	double c0,c1,c2,c3,d0,d1,d2,d3,e0,e1,e2,e3,max;

	// initialize 
	for(i=0;i<(*n);i++) {
		x0[i] = 0;
		x1[i] = 0;
		x2[i] = 0;
		x3[i] = 0;
	}
	
    //forward elimination
	for(k=0; k<(*n)-1; k++) {
		
		//pivotting
		max = fabs(a0[(*n)*k+k]);
		max_i = k;
		for(i=k+1;i<(*n);i++) {
			if(fabs(a0[(*n)*k+i]) > max) {
				max = fabs(a0[(*n)*k+i]);
				max_i = i;
			}
		}

		//swap
		for(j=0;j<(*n);j++) {
			c0 = a0[(*n)*j+k];	
			c1 = a1[(*n)*j+k];	
			c2 = a2[(*n)*j+k];
			c3 = a3[(*n)*j+k];
			a0[(*n)*j+k] = a0[(*n)*j+max_i];	
			a1[(*n)*j+k] = a1[(*n)*j+max_i];
			a2[(*n)*j+k] = a2[(*n)*j+max_i];
			a3[(*n)*j+k] = a3[(*n)*j+max_i];
			a0[(*n)*j+max_i] = c0;
			a1[(*n)*j+max_i] = c1;
			a2[(*n)*j+max_i] = c2;	
			a3[(*n)*j+max_i] = c3;
		}
		c0 = b0[k];
		c1 = b1[k];
		c2 = b2[k];	
		c3 = b3[k];
		b0[k] = b0[max_i];	
		b1[k] = b1[max_i];
		b2[k] = b2[max_i];	
		b3[k] = b3[max_i];
		b0[max_i] = c0;		
		b1[max_i] = c1;		
		b2[max_i] = c2;		
		b3[max_i] = c3;
		for(i=k+1; i<(*n); i++) {
			d0 = 0; d1 = 0;	d2 = 0; d3 = 0;
			//d = a[i][k] / a[k][k];
			//division
			qd_div_qd(&d0, &d1, &d2, &d3, a0[(*n)*k+i], a1[(*n)*k+i], a2[(*n)*k+i], a3[(*n)*k+i], a0[(*n)*k+k], a1[(*n)*k+k], a2[(*n)*k+k], a3[(*n)*k+k]);
			
			for(j=k+1; j<(*n); j++) {
				c0 = 0; c1 = 0;	c2 = 0; c3 = 0;
				e0 = 0; e1 = 0;	e2 = 0; e3 = 0;
				//a[i][j] = a[i][j] - ( d * a[k][j] );
				//multiplication
				qd_mul_qd(&c0, &c1, &c2, &c3, d0, d1, d2, d3, a0[(*n)*j+k], a1[(*n)*j+k], a2[(*n)*j+k], a3[(*n)*j+k]);
				//subtraction
				qd_sub_qd(&e0, &e1, &e2, &e3, a0[(*n)*j+i], a1[(*n)*j+i], a2[(*n)*j+i], a3[(*n)*j+i], c0, c1, c2, c3);
				a0[(*n)*j+i] = e0; a1[(*n)*j+i] = e1; a2[(*n)*j+i] = e2; a3[(*n)*j+i] = e3;
			}
			c0 = 0; c1 = 0;	c2 = 0; c3 = 0;
			e0 = 0; e1 = 0;	e2 = 0; e3 = 0;
			//b[i] = b[i] - ( d * b[k] );
			qd_mul_qd(&c0, &c1, &c2, &c3, d0, d1, d2, d3, b0[k], b1[k], b2[k], b3[k]);
			qd_sub_qd(&e0, &e1, &e2, &e3, b0[i], b1[i], b2[i], b3[i], c0, c1, c2, c3);
			b0[i] = e0;	b1[i] = e1;	b2[i] = e2;	b3[i] = e3;
		}
	}

	// backward substitution
	for(i=(*n)-1; i>=0; i--) {
		c0 = b0[i]; c1 = b1[i]; c2 = b2[i]; c3 = b3[i];
		for(j=i+1; j<(*n); j++) {
			//[c,cc] -= a[i][j] * x[j];
			//multiplication
			qd_mul_qd(&d0, &d1, &d2, &d3, a0[(*n)*j+i], a1[(*n)*j+i], a2[(*n)*j+i], a3[(*n)*j+i], x0[j], x1[j], x2[j], x3[j]);
			//subtraction
			qd_sub_qd(&e0, &e1, &e2, &e3, c0, c1, c2, c3, d0, d1, d2, d3);
			c0 = e0; c1 = e1; c2 = e2; c3 = e3;
		}
		//x[i] = (1.0 / A[i][i])*([c,cc]);
		//division
		qd_div_qd(&e0, &e1, &e2, &e3, c0, c1, c2, c3, a0[(*n)*i+i], a1[(*n)*i+i], a2[(*n)*i+i], a3[(*n)*i+i]);
		x0[i] = e0; x1[i] = e1; x2[i] = e2; x3[i] = e3;
	}
	//free(yhi);
	//free(ylo);
	return(0);
}



// LU decomposition
int qdlu(double *l0, double *l1, double *l2, double *l3, double *u0, double *u1, double *u2, double *u3, double *a0, double *a1, double *a2, double *a3,int *n)
{
	int i,j,k;
	double c0,c1,c2,c3,d0,d1,d2,d3;

	// initialize 
	for(i=0;i<(*n);i++) {
		for(j=0;j<(*n);j++) {
			if(i == j) {
				l0[(*n)*j+i] = 1.0;	l1[(*n)*j+i] = 0.0;	l2[(*n)*j+i] = 0.0; l3[(*n)*j+i] = 0.0;
				u0[(*n)*j+i] = 1.0;	u1[(*n)*j+i] = 0.0;	u2[(*n)*j+i] = 0.0; u3[(*n)*j+i] = 0.0;
			} else {
				l0[(*n)*j+i] = 0.0;	l1[(*n)*j+i] = 0.0;	l2[(*n)*j+i] = 0.0; l3[(*n)*j+i] = 0.0;
				u0[(*n)*j+i] = 0.0;	u1[(*n)*j+i] = 0.0;	u2[(*n)*j+i] = 0.0; u3[(*n)*j+i] = 0.0;
			}
		}
		u0[(*n)*i] = a0[(*n)*i]; u1[(*n)*i] = a1[(*n)*i]; u2[(*n)*i] = a2[(*n)*i]; u3[(*n)*i] = a3[(*n)*i];
	}
	
	for(k=0; k<(*n)-1; k++) {
		for(i=k+1; i<(*n); i++) {
			c0 = 0; c1 = 0;	c2 = 0; c3 = 0;
			//l[i][k] = a[i][k] / a[k][k];
			//division
			qd_div_qd(&d0, &d1, &d2, &d3, a0[(*n)*k+i], a1[(*n)*k+i], a2[(*n)*k+i], a3[(*n)*k+i], a0[(*n)*k+k], a1[(*n)*k+k], a2[(*n)*k+k], a3[(*n)*k+k]);
			l0[(*n)*k+i] = d0; l1[(*n)*k+i] = d1; l2[(*n)*k+i] = d2; l3[(*n)*k+i] = d3;
			
			for(j=k; j<(*n); j++) {
				c0 = 0; c1 = 0;	c2 = 0; c3 = 0;
				//a[i][j] = a[i][j] - ( l[i][k] * a[k][j] );
				//[c,cc] = l[i][k] * a[k][j]
				//multiplication
				qd_mul_qd(&c0, &c1, &c2, &c3, l0[(*n)*k+i], l1[(*n)*k+i], l2[(*n)*k+i], l3[(*n)*k+i], a0[(*n)*j+k], a1[(*n)*j+k], a2[(*n)*j+k], a3[(*n)*j+k]);
				//a[i][j] = a[i][j] - [c,cc];
				//subtraction			
				qd_sub_qd(&d0, &d1, &d2, &d3, a0[(*n)*j+i], a1[(*n)*j+i], a2[(*n)*j+i], a3[(*n)*j+i], c0, c1, c2, c3);
				a0[(*n)*j+i] = d0; a1[(*n)*j+i] = d1; a2[(*n)*j+i] = d2; a3[(*n)*j+i] = d3;
				u0[(*n)*j+i] = d0; u1[(*n)*j+i] = d1; u2[(*n)*j+i] = d2; u3[(*n)*j+i] = d3;
			}
		}
	}
	return(0);
}



/**********************************/
/**  compute inverse matrix of a  */
/**********************************/
int qdinv(double *x0, double *x1, double *x2, double *x3, double *a0, double *a1, double *a2, double *a3, double *y0, double *y1, double *y2, double *y3, int *n)
{
	int i,j,k;
	double c0,c1,c2,c3,d0,d1,d2,d3,e0,e1,e2,e3;
	//double *yhi,*ylo;
	//yhi = (double *)malloc(sizeof(double) * (*n));
	//ylo = (double *)malloc(sizeof(double) * (*n));
    
	// initialize 
	for(i=0;i<(*n);i++) {
		for(j=0;j<(*n);j++) {
			if(i == j) {
				x0[(*n)*j+i] = 1.0; x1[(*n)*j+i] = 0.0; x2[(*n)*j+i] = 0.0; x3[(*n)*j+i] = 0.0;
			} else {
				x0[(*n)*j+i] = 0.0; x1[(*n)*j+i] = 0.0; x2[(*n)*j+i] = 0.0; x3[(*n)*j+i] = 0.0;
			}
		}
		y0[i] = 0; y1[i] = 0; y2[i] = 0; y3[i] = 0;
	}
	
	// LU decomposition
	for(k=0; k<(*n)-1; k++) {
		for(i=k+1; i<(*n); i++) {
			//a[i][k] = a[i][k] / a[k][k];
			//division
			qd_div_qd(&d0, &d1, &d2, &d3, a0[(*n)*k+i], a1[(*n)*k+i], a2[(*n)*k+i], a3[(*n)*k+i], a0[(*n)*k+k], a1[(*n)*k+k], a2[(*n)*k+k], a3[(*n)*k+k]);
			a0[(*n)*k+i] = d0; a1[(*n)*k+i] = d1; a2[(*n)*k+i] = d2; a3[(*n)*k+i] = d3;
			
			for(j=k+1; j<(*n); j++) {
				//a[i][j] = a[i][j] - ( a[i][k] * a[k][j] );
				//*[c,cc] = a[i][k] * a[k][j]
				//multiplication
				qd_mul_qd(&c0, &c1, &c2, &c3, a0[(*n)*k+i], a1[(*n)*k+i], a2[(*n)*k+i], a3[(*n)*k+i], a0[(*n)*j+k], a1[(*n)*j+k], a2[(*n)*j+k], a3[(*n)*j+k]);
				//*a[i][j] = a[i][j] - [c,cc];
				//subtraction			
				qd_sub_qd(&d0, &d1, &d2, &d3, a0[(*n)*j+i], a1[(*n)*j+i], a2[(*n)*j+i], a3[(*n)*j+i], c0, c1, c2, c3);
				a0[(*n)*j+i] = d0; a1[(*n)*j+i] = d1; a2[(*n)*j+i] = d2; a3[(*n)*j+i] = d3;
			}
		}
	}
    
	for(i=0; i<(*n); i++) {
		y0[0] = x0[(*n)*i];
		y1[0] = x1[(*n)*i];
		y2[0] = x2[(*n)*i];
		y3[0] = x3[(*n)*i];
		for(k=1; k<(*n); k++) {
			c0 = 0; c1 = 0; c2 = 0; c3 = 0;
			for(j=0; j<k; j++) {
				//[c,cc] += a[k][j] * y[j];
				//multiplication
				qd_mul_qd(&d0, &d1, &d2, &d3, a0[(*n)*j+k], a1[(*n)*j+k], a2[(*n)*j+k], a3[(*n)*j+k], y0[j], y1[j], y2[i], y3[i]);
				//addition
				qd_add_qd(&e0, &e1, &e2, &e3, c0, c1, c2, c3, d0, d1, d2, d3);
				c0 = e0; c1= e1; c2 = e2; c3= e3;
			}
			//y[k] = x[k][i] - [c,cc];
			//subtraction
			qd_sub_qd(&e0, &e1, &e2, &e3, x0[(*n)*i+k], x1[(*n)*i+k], x2[(*n)*i+k], x3[(*n)*i+k], c0, c1, c2, c3);
			y0[k] = e0; y1[k] = e1; y2[k] = e2; y3[k] = e3;
		}
        
		for(k=(*n)-1; k>=0; k--) {
			c0 = 0; c1 = 0; c2 = 0; c3 = 0;
			for(j=k+1; j<(*n); j++) {
				//[c,cc] += a[k][j] * x[j][i];
				//multiplication
				qd_mul_qd(&d0, &d1, &d2, &d3, a0[(*n)*j+k], a1[(*n)*j+k], a2[(*n)*j+k], a3[(*n)*j+k], x0[(*n)*i+j], x1[(*n)*i+j], x2[(*n)*i+j], x3[(*n)*i+j]);
				//addition
				qd_add_qd(&e0, &e1, &e2, &e3, c0, c1, c2, c3, d0, d1, d2, d3);
				c0 = e0; c1 = e1; c2 = e2; c3 = e3;
			}
			//x[k][i] = (1.0 / A[k][k])*(y[k] - [c,cc]);
			//subtraction
			qd_sub_qd(&e0, &e1, &e2, &e3, y0[k], y1[k], y2[k], y3[k], c0, c1, c2, c3);
			x0[(*n)*i+k] = e0; x1[(*n)*i+k] = e1; x2[(*n)*i+k] = e2; x3[(*n)*i+k] = e3;
			//division
			qd_div_qd(&e0, &e1, &e2, &e3, x0[(*n)*i+k], x1[(*n)*i+k], x2[(*n)*i+k], x3[(*n)*i+k], a0[(*n)*k+k], a1[(*n)*k+k], a2[(*n)*k+k], a3[(*n)*k+k]);
			x0[(*n)*i+k] = e0; x1[(*n)*i+k] = e1; x2[(*n)*i+k] = e2; x3[(*n)*i+k] = e3;
		}
	}
	//free(yhi);
	//free(ylo);
	return(0);
}


/************************/
/**  inner production  **/
/************************/
int qdip(double *c, double *x0, double *x1, double *x2, double *x3, double *y0, double *y1, double *y2, double *y3, int *m)
{
	c[0] = 0.0;
	c[1] = 0.0;
	c[2] = 0.0;
	c[3] = 0.0;
	int i;
	double x,y,w,z,p,q,r,s;
    
	for (i = 0 ; i < *m ; i++) {
		//multiplication
		qd_mul_qd(&x, &y, &w, &z, x0[i], x1[i], x2[i], x3[i], y0[i], y1[i], y2[i], y3[i]);
		//addition
		qd_add_qd(&p, &q, &r, &s, c[0], c[1], c[2], c[3], x, y, w, z);
		c[0] = p;	c[1] = q;	c[2] = r;	c[3] = s;
	}
	return(0);
}


/***********************/
/**  vector two norm  **/
/***********************/
int qdnorm_two(double *c, double *x0, double *x1, double *x2, double *x3, int *m)
{
	c[0] = 0.0;	c[1] = 0.0;	c[2] = 0.0;	c[3] = 0.0;
	int i;
	double p,q,r,s,x,y,w,z;
    
	qd_sqr(&x, &y, &w, &z, x0[0], x1[0], x2[0], x3[0]);
	c[0] = x; c[1] = y;	c[2] = w;	c[3] = z;
	
	for (i = 1 ; i < *m ; i++) {
		qd_sqr(&x, &y, &w, &z, x0[i], x1[i], x2[i], x3[i]);
		qd_add_qd(&p, &q, &r, &s, c[0], c[1], c[2], c[3], x, y, w, z);
		c[0] = p; c[1] = q;	c[2] = r;	c[3] = s;
	}
	qdsqrt_scl(&p, &q, &r, &s, c[0], c[1], c[2], c[3]);
	c[0] = p; c[1] = q;	c[2] = r;	c[3] = s;
	return(0);
}

/*****************/
/**  summation  **/
/*****************/
int qdsum(double *c, double *x0, double *x1, double *x2, double *x3, int *m, int *n)
{
	c[0] = 0.0;	c[1] = 0.0;	c[2] = 0.0;	c[3] = 0.0;
	int i;
	double x,y,w,z;
    
	c[0] = x0[0]; c[1] = x1[0];	c[2] = x2[0];	c[3] = x3[0];
	
	for (i = 1 ; i < (*m)*(*n) ; i++) {
		qd_add_qd(&x, &y, &w, &z, c[0], c[1], c[2], c[3], x0[i], x1[i], x2[i], x3[i]);
		c[0] = x; c[1] = y;	c[2] = w;	c[3] = z;
	}
	return(0);
}


// qd addition
// qd + double
inline void qd_add_s(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b)
{
	double e,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;
    
	two_sum(&x, &y, a0, b);	c0[0] = x;	e = y;
	two_sum(&x, &y, a1, e);		c1[0] = x;	e = y;
	two_sum(&x, &y, a2, e);		c2[0] = x;	e = y;
	two_sum(&x, &y, a3, e);		c3[0] = x;	e = y;
	renorm(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	
    return;
}

// qd addition
// qd + qd
inline void qd_add_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3)
{
	double e1,e2,e3,e4,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;

	two_sum(&x, &y, a0, b0);	c0[0] = x;	e1 = y;
	two_sum(&x, &y, a1, b1);	c1[0] = x;	e2 = y;
	two_sum(&x, &y, c1[0], e1);	c1[0] = x;	e1 = y;
	two_sum(&x, &y, a2, b2);	c2[0] = x;	e3 = y;
  	three_sum(&x, &y, &z, c2[0], e2, e1);	c2[0] = x;	e1 = y;	e2 = z;
	two_sum(&x, &y, a3, b3);	c3[0] = x;	e4 = y;
	three_sum2(&x, &y, c3[0], e3, e1);		c3[0] = x;	e1 = y;
	e1 = e1 + e2 + e4;
	renorm(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e1);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}

// qd subtraction
// qd - qd
inline void qd_sub_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3)
{
	double e1,e2,e3,e4,x,y,w,z;
	b0 = -b0; b1 = -b1;	b2 = -b2; b3 = -b3;	
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;

	two_sum(&x, &y, a0, b0);	c0[0] = x;	e1 = y;
	two_sum(&x, &y, a1, b1);	c1[0] = x;	e2 = y;
	two_sum(&x, &y, c1[0], e1);	c1[0] = x;	e1 = y;
	two_sum(&x, &y, a2, b2);	c2[0] = x;	e3 = y;
  	three_sum(&x, &y, &z, c2[0], e2, e1);	c2[0] = x;	e1 = y;	e2 = z;
	two_sum(&x, &y, a3, b3);	c3[0] = x;	e4 = y;
	three_sum2(&x, &y, c3[0], e3, e1);		c3[0] = x;	e1 = y;
	e1 = e1 + e2 + e4;
	renorm(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e1);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}

// qd multiplication
// double * qd
inline void s_mul_qd(double *c0, double *c1, double *c2, double *c3, double b, double a0, double a1, double a2, double a3)
{
	double e0,e1,e2,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;
		
	two_prod(&x, &y, a0, b);	c0[0] = x;	e0 = y;
	two_prod(&x, &y, a1, b);	c1[0] = x;	e1 = y;
	two_sum(&x, &y, c1[0], e0);	c1[0] = x;	e0 = y;
	two_prod(&x, &y, a2, b);	c2[0] = x;	e2 = y;
	three_sum(&x, &y, &z, c2[0], e1, e0);	c2[0] = x;	e0 = y;	e1 = z;
	c3[0] = a3*b;
	three_sum2(&x, &y, c3[0], e2, e0);		c3[0] = x;	e0 = y;
	e0 = e0 + e1;
	
	renorm(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e0);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}

// qd multiplication
// qd(scalar) * qd(scalar)
inline void qd_mul_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3)
{
	double p10,p01,p11,p20,p02,e00,e10,e01,e11,e20,e02,x,y,w,z;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;
		
	//O(1) terms
	two_prod(&x, &y, a0, b0);
	c0[0] = x;	//O(1) terms
	e00 = y;

	//O(eps) terms
	two_prod(&x, &y, a0, b1);
	p01 = x;
	e01 = y;

	two_prod(&x, &y, a1, b0);
	p10 = x;
	e10 = y;

	three_sum(&x, &y, &z, p01, p10, e00);
	c1[0] = x;	//O(eps)
	p10 = y;	//O(eps^2)
	p01 = z;	//O(eps^3)

	//O(eps^2) terms
	two_prod(&x, &y, a0, b2);
	p02 = x;
	e02 = y;

	two_prod(&x, &y, a1, b1);
	p11 = x;
	e11 = y;

	two_prod(&x, &y, a2, b0);
	p20 = x;
	e20 = y;

	//six three sum for p10, e01, e10, p02, p11, p20
	three_sum(&x, &y, &z, p10, e01, e10);
	p10 = x;
	e01 = y;
	e10 = z;

	three_sum(&x, &y, &z, p02, p11, p20);
	p02 = x;
	p11 = y;
	p20 = z;

	two_sum(&x, &y, p02, p10);
	c2[0] = x;	//O(eps^2) terms
	p10 = y;

	two_sum(&x, &y, p11, e01);
	p11 = x;
	e01 = y;

	two_sum(&x, &y, p10, p11);
	p10 = x;	//O(eps^3) terms
	p11 = y;

	e10 = e10 + p20 + e01 + p11;	//O(eps^4) terms
		
	//O(eps^3) terms
	c3[0] = p10 + a0*b3 + a1*b2 + a2*b1 + a3*b0 + e02 + e11 + e20;

	renorm(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e10);
	c0[0] = x;
	c1[0] = y;
	c2[0] = w;
	c3[0] = z;
	return;
}

//qd division
//qd / qd
inline void qd_div_qd(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3, double b0, double b1, double b2, double b3)
{
	double x,y,w,z,t0,t1,t2,t3,r0,r1,r2,r3,e;
	
		c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;

		c0[0] = a0/b0;
		// reminder a - c_0*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c0[0], b0, b1, b2, b3);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (a - t)
		qd_add_qd(&x, &y, &w, &z, a0, a1, a2, a3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c1[0] = r0/b0;
		// reminder r - c_1*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c1[0], b0, b1, b2, b3);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c2[0] = r0/b0;
		// reminder r - c_2*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c2[0], b0, b1, b2, b3);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;

		c3[0] = r0/b0;
		// reminder r - c_3*b
		//multiplication
		s_mul_qd(&x, &y, &w, &z, c3[0], b0, b1, b2, b3);
		t0 = -x;	t1 = -y;	t2 = -w;	t3 = -z;
		//qd subtruction (r - t)
		qd_add_qd(&x, &y, &w, &z, r0, r1, r2, r3, t0, t1, t2, t3);
		r0 = x;	r1 = y;	r2 = w;	r3 = z;
  
		e = r0/b0;
		renorm(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], e);
		c0[0] = x;
		c1[0] = y;
		c2[0] = w;
		c3[0] = z;
  return;
}



inline void fast_two_sum(double *s, double *e, double a, double b)
{
	double v;
	s[0] = 0.0; e[0] = 0.0;

	s[0] = a + b;
	v = s[0] - a;
	e[0] = b - v;
	
	return;
}

inline void two_sum(double *s, double *e, double a, double b)
{
	double v;
	s[0] = 0.0; e[0] = 0.0;

	s[0] = a + b;
	v = s[0] - a;
	e[0] = (a - (s[0] - v)) + (b - v);
	
	return;
}

inline void three_sum(double *d, double *e, double *f, double a, double b, double c)
{
	double t1,t2,t3,v;
	d[0] = 0.0; e[0] = 0.0; f[0] = 0.0;

	t1= a + b;
	v = t1 - a;
	t2= (a - (t1 - v))+(b - v);
	
	d[0] = t1 + c;
	v = d[0] - t1;
	t3= (t1 - (d[0] - v))+(c - v);
	
	e[0] = t2 + t3;
	v = e[0] - t2;
	f[0] = (t2 - (e[0] - v))+(t3 - v);
	
	return;
}

inline void three_sum2(double *d, double *e, double a, double b, double c)
{
	double t1,t2,t3,v;
	d[0] = 0.0; e[0] = 0.0;

	t1= a + b;
	v = t1 - a;
	t2= (a - (t1 - v))+(b - v);
  
	d[0] = t1 + c;
	v = d[0] - t1;
	t3= (t1 - (d[0] - v))+(c - v);
  
	e[0] = t2 + t3;
	
	return;
}

inline void two_prod(double *p, double *e, double a, double b)
{
	double t,ah,al,bh,bl;

	p[0] = a * b;

	t = 134217729 * a;
	ah = t -(t - a);
	al = a - ah;
	t = 134217729 * b;
	bh = t -(t - b);
	bl = b - bh;

	e[0] = ((ah*bh - p[0]) + ah*bl + al*bh) + al*bl;

	return;
}

inline void sqr(double *p, double *e, double a)
{
	double t,ah,al;
    
	p[0] = a * a;
	t = 134217729 * a;
	ah = t -(t - a);
	al = a - ah;
	e[0] = ((ah*ah - p[0]) + (ah*al)*2.0) + al*al;
    
	return;
}

inline void renorm(double *b0, double *b1, double *b2, double *b3, double a0, double a1, double a2, double a3, double a4)
{
	double t0,t1,t2,t3,t4,s,ss;
	s = 0.0; ss = 0.0;
	t0 = 0.0;t1 = 0.0;t2 = 0.0;t3 = 0.0;t4 = 0.0;
	/*
	fast_two_sum(&x, &y, a3, a4);
	s = x;
	t4 = y;
	fast_two_sum(&x, &y, a2, s);
	s = x;
	t3 = y;
	fast_two_sum(&x, &y, a1, s);
	s = x;
	t2 = y;
	fast_two_sum(&x, &y, a0, s);
	t0 = x;
	t1 = y;
	if(t1 != 0.0) {
		fast_two_sum(&x, &y, t1, t2);
		t1 = x;
		t2 = y;
		if(t2 != 0.0) {
			fast_two_sum(&x, &y,t2, t3);
			t2 = x;
			t3 = y;
			if(t3 != 0.0) {
				t3 += t4;
			} else {
				t2 += t4;
			}
		} else {
			fast_two_sum(&x, &y, t1, t3);
			t1 = x;
			t2 = y;
			if(t2 != 0.0) {
				fast_two_sum(&x, &y, t2, t4);
				t2 = x;
				t3 = y;
			} else {
				fast_two_sum(&x, &y, t1, t4);
				t1 = x;
				t2 = y;
			}
		}
	} else {
		fast_two_sum(&x, &y, t0, t2);
		t0 = x;
		t1 = y;
		if(t1 != 0.0) {
			fast_two_sum(&x, &y, t1, t3);
			t1 = x;
			t2 = y;
			if(t2 != 0.0) {
				fast_two_sum(&x, &y, t2, t4);
				t2 = x;
				t3 = y;
			} else {
				fast_two_sum(&x, &y, t1, t4);
				t1 = x;
				t2 = y;
			}
		} else {
			fast_two_sum(&x, &y, t0, t3);
			t0 = x;
			t1 = y;
			if(t1 != 0.0) {
				fast_two_sum(&x, &y, t1, t4);
				t1 = x;
				t2 = y;
			} else {
				fast_two_sum(&x, &y, t0, t4);
				t0 = x;
				t1 = y;
			}
		}
	}
	*/
	//[s,t4] = fast_two_sum(a4,a5);
	s = a3 + a4;
	t3 = a4 - (s - a3);
	//[ss,t3] = fast_two_sum(a3,s);
	ss = a2 + s;
	t2 = s - (ss - a2);
	//[s,t2] = fast_two_sum(a2,ss);
	s  = a1 + ss;
	t1 = ss - (s - a1);
	//[b1,t1] = fast_two_sum(a1,s);
	b0[0] = a0 + s;
	t0 = s - (b0[0] - a0);
	//[s,t3] = fast_two_sum(t3,t4);
	s = t2 + t3;
	t2 = t3 - (s - t2);
	//[ss,t2] = fast_two_sum(t2,s);
	ss = t1 + s;
	t1 = s - (ss - t1);
	//[b2,t1] = fast_two_sum(t1,ss);
	b1[0] = t0 + ss;
	t0 = ss - (b1[0] - t0);
	//[s,t2] = fast_two_sum(t2,t3);
	s = t1 + t2;
	t1 = t2 - (s -t1);
	//[b3,t1] = fast_two_sum(t1,s);
	b2[0] = t0 + s;
	t0 = s - (b2[0] - t0);
	b3[0] = t0 + t1;
	return;
}


inline void qd_sqr(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3)
{
	double p01,p11,p02,p03,p12,e00,e01,e11,e02,x,y,w,z,s0,s1,t0,t1;
	c0[0] = 0.0; c1[0] = 0.0; c2[0] = 0.0; c3[0] = 0.0;
    
	//O(1) term
	sqr(&x, &y, a0);			c0[0] = x;	e00 = y;	//O(1) term ok
    
	//O(eps) term
	two_prod(&x, &y, a0, a1);	p01 = 2.0*x;	e01 = 2.0*y;
    
	//O(eps^2) terms
	two_prod(&x, &y, a0, a2);	p02 = 2.0*x;	e02 = 2.0*y;
	sqr(&x, &y, a1);			p11 = x;		e11 = y;
    
	two_sum(&x, &y, p01, e00);	c1[0] = x;	e00 = y;		//O(eps) term ok
	two_sum(&x, &y, e00, e01);	e00 = x;	e01 = y;
	two_sum(&x, &y, p02, p11);	p02 = x;	p11 = y;
	two_sum(&x, &y, e00, p02);	s0 = x;		t0 = y;
	two_sum(&x, &y, e01, p11);	s1 = x;		t1 = y;
	two_sum(&x, &y, s1, t0);	s1 = x;		t0 = y;
    
	t0 = t0 + t1;
    
	fast_two_sum(&x, &y, s1, t0);	s1 = x;		t0 = y;
	fast_two_sum(&x, &y, s0, s1);	c2[0] = x;	t1 = y;		//O(eps^2) term ok
	fast_two_sum(&x, &y, t1, t0);	p11 = x;	e00 = y;
    
	//O(eps^3) terms
	p03 = 2.0 * a0 * a3;
	p12 = 2.0 * a1 * a2;
    
	two_sum(&x, &y, p03, p12);	p03 = x;	p12 = y;
	two_sum(&x, &y, e02, e11);	e02 = x;	e11 = y;
	two_sum(&x, &y, p03, e02);	t0 = x;		t1 = y;
    
	t1 = t1 + p12 + e11;
    
	two_sum(&x, &y, p11, t0);	c3[0] = x;	p03 = y;		//O(eps^3) term ok
	p03 = p03 + e00 + t1;									//O(eps^4) term ok
    
	renorm(&x, &y, &w, &z, c0[0], c1[0], c2[0], c3[0], p03);
	c0[0] = x;	c1[0] = y;	c2[0] = w;	c3[0] = z;
	return;
}

inline void qdsqrt_scl(double *c0, double *c1, double *c2, double *c3, double a0, double a1, double a2, double a3)
{
	double h0,h1,h2,h3,x0,x1,x2,x3,p,q,r,s;
	
	c0[0] = 0.0;	c1[0] = 0.0;	c2[0] = 0.0;	c3[0] = 0.0;
    
	c0[0] = 1.0/sqrt(a0);
	h0 = 0.5*a0;	h1 = 0.5*a1;	h2 = 0.5*a2;	h3 = 0.5*a3;	
    
	qd_sqr(&x0, &x1, &x2, &x3, c0[0], c1[0], c2[0], c3[0]);					//x_k^2
	qd_mul_qd(&p, &q, &r, &s, h0, h1, h2, h3, x0, x1, x2, x3);				//0.5 * a * x_k^2
	x0 = -p;	x1 = -q;	x2 = -r;	x3 = -s;
	qd_add_s(&p, &q, &r, &s, x0, x1, x2, x3, 0.5);							//0.5 - 0.5 * a * x_k^2
	qd_mul_qd(&x0, &x1, &x2, &x3, p, q, r, s, c0[0], c1[0], c2[0], c3[0]);	//(0.5 - 0.5 * a * x_k^2)*x_k
	qd_add_qd(&p, &q, &r, &s, c0[0], c1[0], c2[0], c3[0], x0, x1, x2, x3);	//x_k+1 = x_k + (0.5 - 0.5 * a * x_k^2)*x_k
	c0[0] = p; c1[0] = q; c2[0] = r; c3[0] = s;
    
	qd_sqr(&x0, &x1, &x2, &x3, c0[0], c1[0], c2[0], c3[0]);					//x_k^2
	qd_mul_qd(&p, &q, &r, &s, h0, h1, h2, h3, x0, x1, x2, x3);				//0.5 * a * x_k^2
	x0 = -p;	x1 = -q;	x2 = -r;	x3 = -s;
	qd_add_s(&p, &q, &r, &s, x0, x1, x2, x3, 0.5);							//0.5 - 0.5 * a * x_k^2
	qd_mul_qd(&x0, &x1, &x2, &x3, p, q, r, s, c0[0], c1[0], c2[0], c3[0]);	//(0.5 - 0.5 * a * x_k^2)*x_k
	qd_add_qd(&p, &q, &r, &s, c0[0], c1[0], c2[0], c3[0], x0, x1, x2, x3);	//x_k+1 = x_k + (0.5 - 0.5 * a * x_k^2)*x_k
	c0[0] = p; c1[0] = q; c2[0] = r; c3[0] = s;
    
	qd_sqr(&x0, &x1, &x2, &x3, c0[0], c1[0], c2[0], c3[0]);					//x_k^2
	qd_mul_qd(&p, &q, &r, &s, h0, h1, h2, h3, x0, x1, x2, x3);				//0.5 * a * x_k^2
	x0 = -p;	x1 = -q;	x2 = -r;	x3 = -s;
	qd_add_s(&p, &q, &r, &s, x0, x1, x2, x3, 0.5);							//0.5 - 0.5 * a * x_k^2
	qd_mul_qd(&x0, &x1, &x2, &x3, p, q, r, s, c0[0], c1[0], c2[0], c3[0]);	//(0.5 - 0.5 * a * x_k^2)*x_k
	qd_add_qd(&p, &q, &r, &s, c0[0], c1[0], c2[0], c3[0], x0, x1, x2, x3);	//x_k+1 = x_k + (0.5 - 0.5 * a * x_k^2)*x_k
    
	qd_mul_qd(&x0, &x1, &x2, &x3, a0, a1, a2, a3, p, q, r, s);	//(0.5 - 0.5 * a * x_k^2)*x_k*a
	c0[0] = x0; c1[0] = x1; c2[0] = x2; c3[0] = x3;
	
	return;
}
