// ====================================================================
// Copyright (C) 2011 - Satoko Kikkawa
//
// This file must be used under the terms of the CeCILL.
// This source file is licensed as described in the file COPYING, which
// you should have received as part of this distribution.  The terms
// are also available at
// http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
// ====================================================================

// *****************************************************************
// QR decomposition with modified Gram-Schmidt orthonormalization
//
// Calling Sequence :
// qdgqr(A)
//
// Arguments :
// A : QD matrix
//
// Return : 
// Q : m,n orthogonal matrix
// R : n~n upper triangular matrix
//
// Authors :
// Copyright (C) 2011 - Satoko Kikkawa
// last updating : 
// *****************************************************************
function [Q,R] = qdgqr(A)
  [m,n]=size(A.d);
  Q = A;
  R = tlist(['qd','d'],zeros(n,n),zeros(n,n),zeros(n,n),zeros(n,n));
  R(1,1) = norm(Q(:,1),2);
  Q(:,1) = Q(:,1) / R(1,1);
  for i=2:n
    q = Q(:,i);
    R(1:i-1,i) = Q(:,1:i-1)'*q;
    for j=1:i-1
      q = q - R(j,i)*Q(:,j);
    end
    R(i,i) = norm(q,2);
    Q(:,i) = q / R(i,i);
  end
endfunction

